package livio.rssreader.backend;
/*
Version 1.0, 16-08-2021, First release by Livio (javalc6@gmail.com)

IMPORTANT NOTICE, please read:

This software is licensed under the terms of the GNU GENERAL PUBLIC LICENSE,
please read the enclosed file license.txt or https://www.gnu.org/licenses/old-licenses/gpl-2.0-standalone.html

Note that this software is freeware and it is not designed, licensed or intended
for use in mission critical, life support and military purposes.

The use of this software is at the risk of the user.
*/
import android.content.Context;
import android.os.Bundle;
import android.speech.tts.TextToSpeech;
import android.speech.tts.UtteranceProgressListener;
import android.util.Log;

import java.util.Locale;

import livio.rssreader.RSSReader;

public final class TTSEngine implements TextToSpeech.OnInitListener {//standalone TTS engine
    private static final String tag = "TTSEngine";

    private final TextToSpeech mTts;
    private TtsState tts_play;//speech

    // DESIGN RULE: only checkTTS() can set state equal to TtsState.failed
    public enum TtsState {initiating, ready, failed, closed}

    public static final String utteranceId_oneshot = "oneshot";
    private static final String utteranceId_first = "first";
    private static final String utteranceId_interim = "interim";
    public static final String utteranceId_last = "last";

    private final UtteranceProgressListener mUPL;

    // The constructor will create a TextToSpeech instance.
    public TTSEngine(Context context, UtteranceProgressListener mUPL) {
        mTts = new TextToSpeech(context, this);
        tts_play = TtsState.initiating;
        this.mUPL = mUPL;
    }

    @Override
    // OnInitListener method to receive the TTS engine status
    public void onInit(int status) {
        // status can be either TextToSpeech.SUCCESS or TextToSpeech.ERROR.
        if (status == TextToSpeech.SUCCESS) {
            tts_play = TtsState.ready;
//			new DoTTS().execute(TtsCommandPending.warming);    // warm-up engine in background to avoid UI lagging-->some device have problems with warm-up
            if (mUPL != null) {
                int listenerResult = mTts.setOnUtteranceProgressListener(mUPL);
            }
        } else {
            // Initialization failed.
            tts_play = TtsState.closed;
            Log.w(tag, "onInit: could not initialize TextToSpeech.");
        }
    }

    public void setSpeechRate(float speechRate) {
        mTts.setSpeechRate(speechRate);
    }

    public int speak(final String text, final int queueMode, final Bundle params, final String utteranceId) {
        return mTts.speak(text, queueMode, params, utteranceId);
    }

    public int speak(final String text, final int queueMode, final String utteranceId) {
        return mTts.speak(text, queueMode, null, utteranceId);
    }

    public void speakSegments(String word, String[] segments) {
        String utteranceId = segments.length == 0 ? utteranceId_oneshot : utteranceId_first;
        if (mTts.speak(word, TextToSpeech.QUEUE_FLUSH, null, utteranceId) != TextToSpeech.ERROR) {
            for (int i = 0; i < segments.length; i++)
                if (segments[i].trim().length() > 0) { //twin something to say?
                    utteranceId = i == segments.length - 1 ? utteranceId_last : utteranceId_interim;
                    mTts.speak(segments[i].trim(), TextToSpeech.QUEUE_ADD, null, utteranceId);
                }
        } else Log.i(tag, "speakSegments: Error");
    }

    public boolean speakSegments(String[] segments) {
        String utteranceId = segments.length == 1 ? utteranceId_oneshot : utteranceId_first;
        if (mTts.speak(segments[0].trim(), TextToSpeech.QUEUE_FLUSH, null, utteranceId) != TextToSpeech.ERROR) {
            for (int i = 1; i < segments.length; i++) {
                if (segments[i].trim().length() > 0) { //twin something to say?
                    utteranceId = i == segments.length - 1 ? utteranceId_last : utteranceId_interim;
                    mTts.speak(segments[i].trim(), TextToSpeech.QUEUE_ADD, null, utteranceId);
                }
            }
            return true;
        } else {
            Log.i(tag, "speakSegments.2: Error");
            return false;
        }
    }

    public boolean isSpeaking() {
        return mTts.isSpeaking();
    }

    public int stop() {
        try {
            if ((tts_play == TtsState.ready)||(tts_play == TtsState.failed)) //speech
                return mTts.stop();
        } catch (IllegalStateException e) {
            // Do nothing: TTS engine is already stopped.
        }
        return -1;
    }

    public void shutdown() {
        mTts.shutdown();
        tts_play = TtsState.closed;
    }

    public void setLanguage(final Locale loc) {
        mTts.setLanguage(loc);
    }

    public TtsState getTTS_state() {
        return tts_play;
    }

    private boolean checkFirst(String a, String b, int n) { // n: maximum number of chars to check
        int i = 0;
        while ((i < n) && (i < a.length()) && (i < b.length())) {
            if (a.charAt(i) != b.charAt(i))
                return false;
            i++;
        }
        return true;
    }


    private int setLanguageTTS(String langcode) {
        if (checkFirst(langcode, Locale.getDefault().toString(), 3)) // check only first 3 characters (e.g. en_US --> en_)
            return mTts.setLanguage(Locale.getDefault()); // use default locale if coherent with the locale of the dictionary
        else return mTts.setLanguage(RSSReader.getLocale(langcode)); // use locale of the dictionary
    }

    public boolean checkTTS(String langcode) {
// WARNING: invoked on Background thread: no UI is allowed here!
        if (((tts_play != TtsState.ready)&&(tts_play != TtsState.failed))) return false; //tts correction: check speak failure
        int result;
        try {//there are weird crashes with Samsung S6 and similar devices, running android 6
            result = setLanguageTTS(langcode);
        } catch (java.lang.IllegalArgumentException e) { //retry with different locale - workaround for exceptions reported by Samsung S6 and similar devices, running android 6
            result = mTts.setLanguage(Locale.US);//force locale US
            Log.d(tag, "setLanguage: forcing US locale due to bug in device");
        }
        if (result == TextToSpeech.LANG_MISSING_DATA ||
                result == TextToSpeech.LANG_NOT_SUPPORTED) {
            // Language data is missing or the language is not supported.
            Log.i(tag, "language not supported by TTS: "+langcode);
            tts_play = TtsState.failed;
            return false;
        } else {
            tts_play = TtsState.ready;
            return true;
        }
    }

}
